#!/usr/bin/env python3
from typing import Any

APPEND_TIMESTAMP = True
SKIP_MEMORY_LAYER_CHECK_AT_CLOSE = True
PIXEL_SIZE = 1
DEFAULT_NUMBER = 0
CONTRAST_ENHANCE = True
DEFAULT_LAYER_NAME = "TemporaryLayer"
DEFAULT_LAYER_CRS = "EPSG:4326"
VERBOSE = False

__all__ = [
    "serialise_qgis_layer",
]


def serialise_qgis_layer(qgis_instance_handle: Any, layer: Any) -> None:
    """
      https://qgis.org/pyqgis/3.28/core/QgsJsonUtils.html#qgis.core.QgsJsonUtils.exportAttributes

      https://qgis.org/pyqgis/3.28/core/QgsJsonExporter.html


          asGeometryCollection

    Returns contents of the geometry as a list of geometries

    asJson

    Exports the geometry to a GeoJSON string.

    asMultiPoint

    Returns the contents of the geometry as a multi-point.

    asMultiPolygon

    Returns the contents of the geometry as a multi-polygon.

    asMultiPolyline

    Returns the contents of the geometry as a multi-linestring.

    asPoint

    Returns the contents of the geometry as a 2-dimensional point.

    asPolygon

    Returns the contents of the geometry as a polygon.

    asPolyline

    Returns the contents of the geometry as a polyline.

    asQPointF

    Returns contents of the geometry as a QPointF if wkbType is WKBPoint, otherwise returns a null QPointF.

    asQPolygonF

    Returns contents of the geometry as a QPolygonF.

    asWkb

    Export the geometry to WKB

    asWkt

    Exports the geometry to WKT


    TODO: Figure this

          :param qgis_instance_handle:
          :param layer:
          :return:
    """

    # noinspection PyUnresolvedReferences
    from qgis.core import QgsVectorLayer, QgsFeature, QgsJsonExporter

    # noinspection PyUnresolvedReferences
    import qgis

    assert isinstance(layer, QgsVectorLayer)

    # geom_geojson_rep = layer.asJson(precision=17)
    layer_geojson_rep = QgsJsonExporter(
        layer, precision=17
    )  # Note that geometries will be automatically reprojected to WGS84 to match GeoJSON spec if either the source vector layer or source CRS is set.

    return layer_geojson_rep


if __name__ == "__main__":
    print("A")



#!/usr/bin/env python3

__author__ = "heider"
__doc__ = r"""

           Created on 5/5/22
           """

__all__ = [
    "store_project_setting",
    "read_project_setting",
    "restore_default_project_settings",
]

import logging
from logging import warning
from typing import Any, Mapping, Optional

from jord import PROJECT_NAME, VERBOSE


def restore_default_project_settings(
    defaults: Optional[Mapping] = None, *, project_name: str = PROJECT_NAME
) -> None:
    """

    :param defaults:
    :param project_name:
    :return:
    """
    if defaults is None:
        defaults = {}

    for key, value in defaults.items():
        store_project_setting(key, value, project_name=project_name)


def store_project_setting(
    key: str, value: Any, *, project_name: str = PROJECT_NAME
) -> None:
    """

    :param key:
    :param value:
    :param project_name:
    :return:
    """
    # noinspection PyUnresolvedReferences
    from qgis.core import QgsProject

    qgis_project = QgsProject.instance()

    if isinstance(value, bool):
        qgis_project.writeEntryBool(project_name, key, value)
    elif isinstance(value, float):
        qgis_project.writeEntryDouble(project_name, key, value)
    # elif isinstance(value, int): # DOES NOT EXIST!
    #    qgis_project.writeEntryNum(project_name, key, value)
    else:
        value = str(value)
        qgis_project.writeEntry(project_name, key, value)

    if VERBOSE:
        logging.info(f"Stored in {project_name} settings {key=} {value=}")


def read_project_setting(
    key: str,
    type_hint: type = None,
    *,
    defaults: Mapping = None,
    project_name: str = PROJECT_NAME,
) -> Any:
    """

    :param key:
    :param type_hint:
    :param defaults:
    :param project_name:
    :return:
    """
    # noinspection PyUnresolvedReferences
    from qgis.core import QgsProject

    qgis_project = QgsProject.instance()

    # read values (returns a tuple with the value, and a status boolean
    # which communicates whether the value retrieved could be converted to
    # its type, in these cases a string, an integer, a double and a boolean
    # respectively)

    if defaults is None:
        defaults = {}

    if type_hint is not None:
        if type_hint is bool:
            val, type_conversion_ok = qgis_project.readBoolEntry(
                project_name, key, defaults.get(key, None)
            )
        elif type_hint is float:
            val, type_conversion_ok = qgis_project.readDoubleEntry(
                project_name, key, defaults.get(key, None)
            )
        elif type_hint is int:
            val, type_conversion_ok = qgis_project.readNumEntry(
                project_name, key, defaults.get(key, None)
            )
        else:
            val, type_conversion_ok = qgis_project.readEntry(
                project_name, key, str(defaults.get(key, None))
            )
    else:
        val, type_conversion_ok = qgis_project.readEntry(
            project_name, key, str(defaults.get(key, None))
        )

    if type_hint is not None:
        val = type_hint(val)

    if False:
        if not type_conversion_ok:
            warning(f"read_plugin_setting: {key} {val} {type_conversion_ok}")

    return val



import logging
import uuid

__all__ = ["randomize_field", "randomize_sub_tree_field"]

from typing import Collection, Union, Any

logger = logging.getLogger(__name__)


def randomize_sub_tree_field(
    selected_nodes: Union[Any, Collection[Any]], field_name: str
) -> None:
    # noinspection PyUnresolvedReferences
    from qgis.core import QgsLayerTreeGroup, QgsLayerTreeLayer, QgsLayerTreeNode

    if isinstance(selected_nodes, QgsLayerTreeLayer):
        randomize_field(selected_nodes, field_name=field_name)
    elif isinstance(selected_nodes, QgsLayerTreeGroup):
        randomize_sub_tree_field(selected_nodes.children(), field_name=field_name)
    elif isinstance(selected_nodes, QgsLayerTreeNode):
        if selected_nodes.nodeType() == QgsLayerTreeNode.NodeGroup:
            randomize_sub_tree_field(selected_nodes.children(), field_name=field_name)
        else:
            logger.error(
                f"Node {selected_nodes} was not supported in randomize_sub_tree_field, skipping"
            )
    else:
        if len(selected_nodes) == 0:
            logger.error(
                f"'Number of selected nodes was {len(selected_nodes)}, please supply some"
            )
            return

        for group in iter(selected_nodes):
            if isinstance(group, QgsLayerTreeLayer):
                randomize_field(group, field_name=field_name)
            elif isinstance(group, QgsLayerTreeGroup):
                randomize_sub_tree_field(group.children(), field_name=field_name)
            elif isinstance(group, QgsLayerTreeNode):
                if group.nodeType() == QgsLayerTreeNode.NodeGroup:
                    randomize_sub_tree_field(group.children(), field_name=field_name)
                else:
                    logger.error(
                        f"Node {group} was not supported in randomize_sub_tree_field, skipping"
                    )
            else:
                logger.error(
                    f"Node {group} was not supported in randomize_sub_tree_field, skipping"
                )


def randomize_field(tree_layer: Any, field_name: str) -> None:  #: QgsLayerTreeLayer
    """
        https://qgis.org/pyqgis/3.28/core/QgsVectorLayer.html#qgis.core.QgsVectorLayer.changeAttributeValues

    changeAttributeValues
    fid: int, newValues: Dict[int, Any], oldValues: Dict[int, Any] = {}, skipDefaultValues: bool = False

        :param field_name:
        :param tree_layer:
        :return:
    """
    if tree_layer is None:
        logger.error(f"Tree layer was None")
        return
    # logger.info(f'Randomizing {field_name} in {tree_layer.layer().name()}')

    layer = tree_layer.layer()

    field_idx = layer.fields().indexFromName(field_name)

    if field_idx >= 0:
        layer.startEditing()
        # layer.beginEditCommand(f"Regenerate {field_name}")
        logger.info(
            f"Randomizing {field_name}:{field_idx} in {tree_layer.layer().name()}"
        )

        for i in range(layer.featureCount() + 1):
            layer.changeAttributeValue(i, field_idx, uuid.uuid4().hex)

        # layer.rollBack()
        # layer.endEditCommand()
        layer.commitChanges()
    else:
        logger.error(f"Did not find {field_name} in {layer.name()}")



#!/usr/bin/env python3
from typing import List, Iterable, Union, Tuple

import numpy
import pyproj
from shapely.geometry.base import BaseGeometry
from shapely.ops import transform

__all__ = [
    "crs_transform_shapely",
    "get_helmert_transformation_parameters",
    "get_affine_transform_parameters",
]


def crs_transform_shapely(
    geoms: Union[BaseGeometry, Iterable[BaseGeometry]],
    from_coordinate_system: str,
    to_coordinate_system: str,
) -> Union[BaseGeometry, List[BaseGeometry]]:
    """
    Project space geometries from one coordinate system to another

    :param geoms: A list of SpacePolygons to project
    :param from_coordinate_system: The source coordinate system
    :param to_coordinate_system: The destination coordinate system
    :return: A list of SpacePolygons projected
    """

    source = pyproj.CRS(from_coordinate_system)
    destination = pyproj.CRS(to_coordinate_system)
    projection = pyproj.Transformer.from_crs(
        source, destination, always_xy=True
    ).transform
    #    project = pyproj.Transformer.from_proj(
    #    pyproj.Proj(init="epsg:4326"),  # source coordinate system
    #    pyproj.Proj(init="epsg:3857"),  # destination coordinate system
    # )

    if isinstance(geoms, Iterable):
        return [transform(projection, geometry) for geometry in geoms]

    return transform(projection, geoms)


def get_helmert_transformation_parameters(
    displacements: Iterable[Tuple[Tuple[float, float], Tuple[float, float]]],
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """
    # Calculate Helmert parameters for 2D transformation
    # https://homepage.univie.ac.at/Wolfgang.Kainz/Lehrveranstaltungen/15th_Nordic_Summer_School/The_Mathematics_of_GIS_Draft.pdf
    # https://riptutorial.com/numpy/example/16034/find-the-least-squares-solution-to-a-linear-system-with-np-linalg-lstsq

    # Displacement lines have to from local coordinate system to Webmercator.
    # The transformation is a 2d helmert and not affine, so the scale factor is the same in x and y direction
    # The angle between x and y axis have is also 90 degree.
    # This would not be the case if target project would have been WGS84

    :param displacements:
    :return:
    """
    A = []
    b = []
    for from_, to_ in displacements:
        from_x, from_y = from_
        to_x, to_y = to_

        A.append([from_x, from_y, 1, 0])
        b.append(to_x)
        A.append([from_y, -from_x, 0, 1])
        b.append(to_y)

    x, residuals, rank, s = numpy.linalg.lstsq(A, b, rcond=-1)

    return x, residuals


def get_affine_transform_parameters(
    displacements: Iterable[Tuple[Tuple[float, float], Tuple[float, float]]]
) -> List[float]:
    """

    :param displacements:
    :return:
    """
    transformation_params, _ = get_helmert_transformation_parameters(displacements)
    a, b, c, d = transformation_params
    return [a, b, -b, a, c, d]
